from os import environ as env

from .Splitter import Splitter

from .Bark import Bark
from .RuTTS import RuTTS
from .SaluteSpeech import SaluteSpeech
from .Crt import Crt
from .Coqui import Coqui
from .Silero import Silero


class RaconteurFactory:
    def __init__(self, gpu: bool = False, ru: bool = False):
        self.gpu = gpu
        self.ru = ru

    def make(self, engine: str, max_n_characters: int = None, artist: str = None, ssml: bool = False):
        match engine:
            case SaluteSpeech.name:
                return SaluteSpeech(
                    # client_id = env['SALUTE_SPEECH_CLIENT_ID'],
                    # client_secret = env['SALUTE_SPEECH_CLIENT_SECRET'],
                    auth = env['SALUTE_SPEECH_AUTH'],
                    artist = 'Nec',
                    splitter = Splitter(4000 if max_n_characters is None else max_n_characters)
                )
            case Bark.name:
                return Bark(
                    artist = 'v2/ru_speaker_6' if self.ru else 'v2/en_speaker_6',
                    splitter = Splitter(200 if max_n_characters is None else max_n_characters)
                )
            case RuTTS.name:
                return RuTTS(
                    artist = 'TeraTTS/natasha-g2p-vits',
                    splitter = Splitter(1000 if max_n_characters is None else max_n_characters),
                    add_time_to_end = 0.1,
                    length_scale = 1.65,
                    gpu = self.gpu
                )
            case Crt.name:
                return Crt(
                    username = env['CRT_USERNAME'],
                    password = env['CRT_PASSWORD'],
                    domain = int(env['CRT_DOMAIN']),
                    artist = 'Vladimir_n',
                    splitter = Splitter(500 if max_n_characters is None else max_n_characters)
                )
            case Coqui.name:
                return Coqui(
                    speaker_wav = 'assets/female.wav',
                    gpu = self.gpu,
                    ru = self.ru,
                    splitter = Splitter(200 if max_n_characters is None else max_n_characters)
                )
            case Silero.name:
                return Silero(
                    model = 'v4' if self.ru else 'v3',
                    gpu = self.gpu,
                    artist = ('xenia' if self.ru else 'en_1') if artist is None else artist,
                    ru = self.ru,
                    splitter = Splitter(400 if max_n_characters is None else max_n_characters),
                    ssml = ssml
                )
            case _:
                raise ValueError(f'Unknown engine {engine}')



from unittest import TestCase, main

from rr import Splitter


class TestSplitter(TestCase):

    def test_no_constraints(self):
        self.assertEqual(Splitter().split('foo'), ('foo', ))

    def test_multiple_tokens_with_single_space(self):
        self.assertEqual(Splitter(max_n_characters = 10).split('foo bar baz qux quux'), ('foo bar', 'baz qux', 'quux'))

    def test_multiple_tokens_with_multiple_intermediate_spaces(self):
        self.assertEqual(Splitter(max_n_characters = 10).split('foo    bar baz qux quux'), ('foo    bar', 'baz qux', 'quux'))

    def test_multiple_tokens_with_multiple_intermediate_spaces_2(self):
        self.assertEqual(Splitter(max_n_characters = 10).split('foo     bar baz qux quux'), ('foo', 'bar baz', 'qux quux'))

    def test_leading_and_trailing_spaces(self):
        self.assertEqual(Splitter(max_n_characters = 10).split('         foo  bar baz qux quux          '), ('foo  bar', 'baz qux', 'quux'))

    def test_long_tokens(self):
        self.assertEqual(Splitter(max_n_characters = 10).split('foo onetwothreefour five sixs eveneight'), ('foo', 'onetwothre', 'efour five', 'sixs', 'eveneight'))


if __name__ == '__main__':
    main()



from os import path, makedirs
from time import time
import math
from pathlib import Path

from click import group, argument, option, Choice
from pandas import read_csv
import numpy as np

# from beep import beep
# from cloud_mail_api import CloudMail
# from fuck import ProfanityHandler
from tqdm import tqdm
# import torch

from .Bark import Bark
from .RuTTS import RuTTS
from .SaluteSpeech import SaluteSpeech
from .Crt import Crt
from .Coqui import Coqui
from .Silero import Silero

from .RaconteurFactory import RaconteurFactory

from .util import one_is_not_none, read  # , drop_accent_marks, drop_empty_lines
from .SpeechIndex import SpeechIndex


ENGINES = Choice((Bark.name, RuTTS.name, SaluteSpeech.name, Crt.name, Coqui.name, Silero.name), case_sensitive = False)


@group()
def main():
    pass


@main.command()
@argument('text', type = str)  # file must be in a format exported by much module: see https://github.com/zeionara/much
@option('--artist-one', '-a1', help = 'first artist to say the replic', default = 'xenia')
@option('--artist-two', '-a2', help = 'second artist to say the replic', default = 'baya')
def alternate(text: str, artist_one: str, artist_two: str):
    assert artist_one != artist_two, 'The two artist must not be same'

    output_path = f'{path.splitext(text)[0]}.mp3'

    factory = RaconteurFactory(gpu = True, ru = True)

    a1 = factory.make(engine = 'silero', artist = artist_one)
    a2 = factory.make(engine = 'silero', artist = artist_two)

    a1_speaks = True

    accumulator = np.array([], dtype = 'float32')

    with open(text, 'r') as file:
        lines = file.read().split('\n')

    pbar = tqdm(total = len(lines))

    for line in lines:
        if line:
            # You can't create two objects for the artist and reuse them - in this case the program hangs after the first iteration
            accumulator = (a1 if a1_speaks else a2).speak(line, save_text = False, accumulator = accumulator)
            a1_speaks = not a1_speaks

        pbar.update()

    a1.save(accumulator, filename = output_path)


@main.command()
@argument('text', type = str, required = False)
@option('--max-n-characters', '-c', help = 'max number of characters given to the speech engine at once', type = int, default = None)
@option('--gpu', '-g', help = 'run model using gpu', is_flag = True)
@option('--engine', '-e', help = 'speaker type to use', type = ENGINES, default = RuTTS.name)
@option('--destination', '-d', help = 'path to the resulting mp3 file', type = str, default = None)
@option('--russian', '-r', help = 'is input text in russian language', is_flag = True)
@option('--txt', '-t', help = 'read text from a plain .txt file located at the given path', type = str, default = None)
@option('--artist', '-a', help = 'speaker id to use for speech generation', type = str, default = None)
@option('--drop-text', '-x', help = 'do not keep source text in generated audio file metadata (for instance, because the text is very long)', is_flag = True)
@option('--batch-size', '-b', help = 'number of characters per generated audio file', type = int, default = None)
@option('--ssml', '-m', help = 'does input text contain ssml tags', is_flag = True)
@option('--first-batch-index', '-f', help = 'in a multibatch setting from what number to start enumerating the batches', type = int, default = 0)
@option('--update', '-u', help = 'update existing files instead of generating new ones', is_flag = True)
def say(
    text: str, max_n_characters: int, gpu: bool, engine: str, destination: str, russian: bool, txt: str, artist: str,
    drop_text: bool, batch_size: int, ssml: bool = False, first_batch_index: int = 0, update: bool = True
):
    match one_is_not_none('Exactly one of input text, path to txt file must be specified', text, txt):
        case 1:
            text = read(txt)

    if batch_size is not None:
        if destination is None:
            txt_stem = Path(txt).stem
            destination = path.join(txt[::-1].split('/', maxsplit = 1)[1][::-1], txt_stem)

            # print(destination)
            # raise ValueError('Destination name is required when splitting output file')

        n_chunks = first_batch_index + math.ceil(len(text) / batch_size)

        # val = input(f'There will be {n_chunks} chunks, ok? (y/N): ')

        # if val != 'y':
        #     return

        if not path.isdir(destination):
            makedirs(destination)

        stem = Path(destination).stem
        batch_index_max_length = len(str(n_chunks))
        # template = "f'" + path.join(destination, f'{stem}-{{batch:0{batch_index_max_length}d}}.mp3') + "'"

        title = f'{stem}-{{batch:0{batch_index_max_length}d}}'
        template = path.join(destination, f'{title}.mp3')

        destination = template

        # batch = 8
        # print(eval(template))

        # print(template.format(batch = 8))

    elif destination is None:
        destination = 'assets/speech.mp3'

    RaconteurFactory(gpu, russian).make(engine, max_n_characters, artist, ssml).speak(
        text, filename = destination, pbar = True, save_text = not drop_text, batch_size = batch_size, first_batch_index = first_batch_index, title = title, update = update
    )


@main.command()
@option('--source', '-s', help = 'path to the input tsv file with anecdotes', type = str, default = 'assets/anecdotes.tsv')
@option('--destination', '-d', help = 'path to the output directory with anecdotes', type = str, default = 'assets/anecdotes')
@option('--max-n-characters', '-c', help = 'max number of characters given to the speech engine at once', type = int, default = None)
@option('--top-n', '-n', help = 'number of entries to handle', type = int, default = None)
@option('--offset', '-o', help = 'number of entries in the beginning to skip', type = int, default = None)
@option('--gpu', '-g', help = 'run model using gpu', is_flag = True)
@option('--engine', '-e', help = 'speaker type to use', type = ENGINES, default = RuTTS.name)
@option('--russian', '-r', help = 'is input text in russian language', is_flag = True)
@option('--skip-if-exists', '-k', help = 'skip anek if audio file with the same name already exists', is_flag = True)
@option('--username', '-u', help = 'cloud mail ru username', type = str)
@option('--password', '-p', help = 'cloud mail ru password', type = str)
@option('--cloud-root', '-x', help = 'root folder where to upload generated mp3 files', type = str)
@option('--upload-and-quit', '-q', help = 'upload files to cloud if they exist before starting speech generation', is_flag = True)
@option('--verbose', '-v', help = 'whether to enable additional logging', is_flag = True)
def handle_aneks(
    source: str, destination: str, max_n_characters: int, top_n: int, offset: int, gpu: bool, engine: str, russian: bool, skip_if_exists: bool,
    username: str, password: str, cloud_root: str, upload_and_quit: bool, verbose: bool
):
    if not path.isdir(destination):
        makedirs(destination)

    df = read_csv(source, sep = '\t')

    n_aneks = 0

    speaker = RaconteurFactory(gpu, russian).make(engine, max_n_characters)

    cm = None

    if username is not None and password is not None and cloud_root is not None:
        cm = CloudMail(username, password)
        cm.auth()

    start = time()

    with beep():
        for _, row in (
            (
                df if offset is None else df.iloc[offset:,]
            )
            if top_n is None else
            (
                df.iloc[:top_n,] if offset is None else df.iloc[offset:top_n,]
            )
        ).loc[:, ('id', 'text', 'source')].iterrows():
            text = row['text']

            name = f'{row["id"]:08d}.{row["source"]}.mp3'
            # name_copy = f'{row["id"]:08d}.{row["source"]} (1).mp3'
            filename = path.join(destination, name)

            # print(f'Handling "{text}"')
            if upload_and_quit and cm is not None and path.isfile(filename):
                # print(filename, f'{cloud_root}/{name}')
                status = None

                while status != 200:
                    response = cm.api.file.add(filename, f'{cloud_root}/{name}')
                    status = response['status']

                # response = cm.api.file(f'{cloud_root}/{name}')

                # if response['status'] != 200:
                #     print(response)

                # cm.api.file.remove(f'{cloud_root}/{name_copy}')

            if not skip_if_exists or not path.isfile(filename):
                if verbose:
                    print(text)

                # try:
                speaker.speak(
                    text = text,
                    filename = filename
                )
                # except Exception:  # on any exception try to repeat again after 10 seconds, there may be a temporary problem with the network
                #     sleep(10)

                #     speaker.speak(
                #         text = text,
                #         filename = filename
                #     )

            n_aneks += 1

            print(f'Handled {n_aneks} aneks')

        elapsed = time() - start
        print(f'Handled {n_aneks} aneks in {elapsed:.5f} seconds ({elapsed / n_aneks:.5f} seconds per anek in average)')


@main.command()
@argument('anecdotes', type = str)
@argument('speech-path', type = str)
@option('--engine', '-e', help = 'speaker type to use', type = ENGINES, default = Crt.name)
@option('--offset', '-o', help = 'number of entries to skip', type = int, default = None)
@option('--cloud-root', '-c', help = 'root folder at the mail.ru cloud - if this option is set, the command works in upload-only mode meaning that existing files are just uploaded to the cloud')
@option('--username', '-u', help = 'username for updating files in the cloud')
@option('--password', '-p', help = 'password for updating files in the cloud')
def uncensor(anecdotes: str, speech_path: str, engine: str, offset: int, cloud_root: str, username: str, password: str):
    cm = None if cloud_root is None else CloudMail(username, password)

    index = SpeechIndex(speech_path)
    ph = ProfanityHandler()

    speaker = RaconteurFactory().make(engine, max_n_characters = 300) if cloud_root is None else None

    df = read_csv(anecdotes, sep = '\t')

    n_rows, _ = df.shape
    pbar = tqdm(total = n_rows, desc = 'Handling documents', initial = 0 if offset is None else offset)

    df = df if offset is None else df.iloc[offset:, ]

    n_spoken = 0

    for i, row in df.iterrows():
        text, changed, _ = ph.uncensor(row['text'])

        if changed:
            if cloud_root is None:
                try:
                    speaker.speak(
                        text = text,
                        filename = index.get(row['source'], row['id']).path
                    )

                    n_spoken += 1
                    pbar.desc = f'Handling documents (spoken: {n_spoken})'
                except Exception:
                    print('Failed to complete the operation, continue from', i)
                    raise
            else:
                location = index.get(row['source'], row['id'])

                # remote_path = f'{cloud_root}/{location.file}'.replace('.mp3', ' (1).mp3')
                remote_path = f'{cloud_root}/{location.file}'

                response = cm.api.file.remove(remote_path)

                # if response['status'] == 200:
                #     print(remote_path)

                if response['status'] != 200:
                    raise ValueError(f'Cannot remove file {location.file}')

                response = cm.api.file.add(location.path, remote_path)

                if response['status'] != 200:
                    raise ValueError(f'Cannot upload file {location.file}')

                print(f'Uploaded file {location.file}')

            # print('=' * 10)
            # print(row['text'])
            # print('-' * 10)
            # print(text)
            # print('*' * 10)
            # print(index.get(row['source'], row['id']))

        pbar.update()


if __name__ == '__main__':
    main()



import re

from RUTTS import TTS
from ruaccent import RUAccent
from transliterate import translit
# from num2words import num2words
# from onnxruntime.capi.onnxruntime_pybind11_state import RuntimeException

from .Raconteur import Raconteur

from .util import translate_numbers


RUTTS_MODEL_PATH = './assets/rutts'
ACCENTIZER_MODEL_PATH = './assets/accentizer'

NUMBER_PATTERN = re.compile('[-+]?[0-9]+')


class RuTTS(Raconteur):
    name = 'rutts'

    def __init__(self, artist: str = 'TeraTTS/natasha-g2p-vits', add_time_to_end = 0.8, length_scale = 1.2, gpu: bool = True, *args, **kwargs):
        self.add_time_to_end = add_time_to_end
        self.length_scale = length_scale
        self.artist = artist

        self.tts = TTS(artist, add_time_to_end = add_time_to_end, save_path = RUTTS_MODEL_PATH, gpu = gpu)
        self.accentizer = accentizer = RUAccent(workdir = ACCENTIZER_MODEL_PATH)
        accentizer.load(omograph_model_size = 'medium')

        super().__init__(*args, **kwargs)

    @property
    def sample_rate(self):
        return 22050

    @property
    def dtype(self):
        return 'float32'

    def predict(self, text):

        # Transliterate

        text = translit(text, 'ru')

        # Replace numbers with words

        text = translate_numbers(text)

        # while (match := NUMBER_PATTERN.search(text)) is not None:
        #     number = match.group(0)
        #     text = text.replace(number, num2words(number, lang = 'ru'), 1)

        # Add accent marks

        try:
            text = self.accentizer.process_all(text)
        except Exception:
            print(f'Can\'t put accents in text "{text}"')

        return self.tts(
            text,
            lenght_scale = self.length_scale
        )

    def set_file_meta(self, file):
        file['artist'] = self.artist



import torch

from transliterate import translit
# from ruaccent import RUAccent

from .Raconteur import Raconteur
from .util import translate_numbers
# from .RuTTS import ACCENTIZER_MODEL_PATH

REPO = 'snakers4/silero-models'
MODEL = 'silero_tts'

DOUBLE_LINE_BREAK = '\n\n'
PAUSE = '\n<break time="2000ms"/>\n'

PRE_PARAGRAPH_PAUSE_LENGTH = 50_000
POST_PARAGRAPH_PAUSE_LENGTH = 50_000


class Silero(Raconteur):
    name = 'silero'

    def __init__(self, model: str = 'v4', gpu: bool = True, artist: str = 'xenia', ru: bool = True, ssml: bool = False, *args, **kwargs):
        self.model = model
        self.artist = artist

        self.language = 'ru' if ru else 'en'

        self.device = torch.device('cuda' if gpu else 'cpu')
        self.ssml = ssml

        # self.accentizer = accentizer = RUAccent(workdir = ACCENTIZER_MODEL_PATH)
        # accentizer.load()

        # model, _ = torch.hub.load(
        #     repo_or_dir = REPO,
        #     model = MODEL,
        #     language = 'ru' if ru else 'en',
        #     speaker = f'{model}_{language}'
        # )

        # self.model = model

        super().__init__(*args, **kwargs)

    def predict(self, text: str):
        # data = self.model.apply_tts(
        #     text = text,
        #     speaker = self.artist,
        #     sample_rate = self.sample_rate
        # )
        model, _ = torch.hub.load(
            repo_or_dir = REPO,
            model = MODEL,
            language = self.language,
            speaker = f'{self.model}_{self.language}'
        )

        model.to(self.device)

        text = translate_numbers(translit(text, 'ru') if self.language == 'ru' else text, lang = self.language)

        if ssml := self.ssml:
            # text_with_tags = self.accentizer.process_all(text)
            # print(text_with_tags)
            # text_with_tags = f'<speak>\n<p>\n{text_with_tags.replace(DOUBLE_LINE_BREAK, PAUSE)}\n</p>\n</speak>'
            text_with_tags = f'<speak>\n<p>\n{text.replace(DOUBLE_LINE_BREAK, PAUSE)}\n</p>\n</speak>'
            text_with_tags = text_with_tags.replace('»', '"').replace('«', '"').replace('&', 'энд')
        else:
            text_with_tags = None

        # if self.ssml:
        #     print(text_with_tags)

        try:
            data = model.apply_tts(
                text = None if ssml else text,
                ssml_text = text_with_tags if ssml else None,
                # text = text,
                speaker = self.artist,
                sample_rate = self.sample_rate,
                put_accent = True,
                put_yo = True
            )
        except Exception:
            print('Text: ', text_with_tags if self.ssml else text)
            raise

        vector = data.numpy()

        if ssml:
            return vector[PRE_PARAGRAPH_PAUSE_LENGTH:-POST_PARAGRAPH_PAUSE_LENGTH]  # cut the pause in the beginning and at the end of paragraph

        return vector

    @property
    def sample_rate(self):
        return 48_000

    @property
    def dtype(self):
        return 'float32'

    def set_file_meta(self, file):
        file['artist'] = self.artist



def one_is_not_none(message: str, *args: str):
    n_args = 0
    last_index = None

    for i, arg in enumerate(args):
        if arg is not None:
            n_args += 1
            last_index = i

    assert n_args == 1, message

    return last_index



import re


SPACE = re.compile('\\s+')


class Splitter:
    def __init__(self, max_n_characters: int = None):
        self.max_n_characters = max_n_characters

    def split(self, text: str):
        if self.max_n_characters is None:
            return (text, )
        if len(text) < 1:
            return ('', )

        max_n_characters = self.max_n_characters

        chunks = []
        n_characters_in_current_chunk = 0
        current_chunk = []

        current_token = []
        current_space = []
        previous_c = None

        n_characters_in_current_token = 0
        n_characters_in_space_after_current_token = 0
        n_characters_in_space_after_previous_token = None
        previous_space = None

        watching_space = True if SPACE.fullmatch(text[0]) else False

        # def increment():
        #     nonlocal n_characters_in_current_token, n_characters_in_space_after_current_token

        #     if watching_space:
        #         n_characters_in_space_after_current_token += 1
        #     else:
        #         n_characters_in_current_token += 1

        def push(c: str):
            nonlocal current_token, current_space, n_characters_in_current_token, n_characters_in_space_after_current_token, previous_space

            if watching_space:
                current_space.append(c)
                n_characters_in_space_after_current_token += 1
            else:
                current_token.append(c)
                n_characters_in_current_token += 1

        def reset(c: str = None):
            nonlocal n_characters_in_current_token, n_characters_in_space_after_current_token, current_token, n_characters_in_current_chunk, n_characters_in_space_after_previous_token, current_chunk
            nonlocal current_space, previous_space

            # print(current_token, current_space, n_characters_in_current_chunk)

            if n_characters_in_current_token > 0:
                if n_characters_in_current_chunk + n_characters_in_current_token + (
                    0 if n_characters_in_space_after_previous_token is None else n_characters_in_space_after_previous_token
                ) <= max_n_characters:
                    n_characters_in_current_chunk += n_characters_in_current_token

                    if n_characters_in_space_after_previous_token is not None:
                        n_characters_in_current_chunk += n_characters_in_space_after_previous_token
                        current_chunk.extend(previous_space)

                    n_characters_in_space_after_previous_token = n_characters_in_space_after_current_token
                    previous_space = current_space

                    current_chunk.extend(current_token)
                else:
                    chunks.append(''.join(current_chunk))

                    current_chunk = current_token
                    n_characters_in_current_chunk = n_characters_in_current_token

                    n_characters_in_space_after_previous_token = n_characters_in_space_after_current_token
                    previous_space = current_space

            # print(current_token, current_chunk)

            if c is not None:
                current_token = [c]
                current_space = []

                n_characters_in_current_token = 1
                n_characters_in_space_after_current_token = 0
            else:
                chunks.append(''.join(current_chunk))

        # unclosed_tag = False

        for c in text:
            # if unclosed_tag or c in ('<', '>'):
            #     if c == '<':
            #         unclosed_tag = True
            #     elif c == '>':
            #         unclosed_tag = False

            #     push(c)
            #     previous_c = c

            #     continue

            if not SPACE.fullmatch(c) and previous_c is not None and SPACE.fullmatch(previous_c):  # token + space has finished
                watching_space = False
                # print(current_token)
                reset(c)
            else:
                if SPACE.fullmatch(c) and previous_c is not None and not SPACE.fullmatch(previous_c):  # token has finished, space has started
                    watching_space = True
                else:
                    if n_characters_in_current_token >= max_n_characters:
                        reset(c)
                        continue
                    # print('--')
                # current_token.append(c)
                push(c)
                # print(n_characters_in_current_chunk, c)

            # n_characters_in_current_chunk += 1
            previous_c = c

            # increment()

        reset()

        return tuple(chunks)



import numpy as np

from TTS.api import TTS

from .Raconteur import Raconteur


class Coqui(Raconteur):
    name = 'coqui'

    def __init__(self, speaker_wav: str, model: str = 'tts_models/multilingual/multi-dataset/xtts_v1', gpu: bool = True, ru: bool = True, *args, **kwargs):
        self.model = model
        self.gpu = gpu
        self.speaker_wav = speaker_wav
        self.ru = ru

        self.tts = TTS(model).to('cpu')  # to('cuda') causes out of memory error

        super().__init__(*args, **kwargs)

    def predict(self, text: str):
        data = self.tts.tts(
            text = text,
            speaker_wav = self.speaker_wav,
            language = 'ru' if self.ru else 'en'
        )

        return np.array(data, dtype = self.dtype)

    @property
    def sample_rate(self):
        return 24_000

    @property
    def dtype(self):
        return 'float32'

    def set_file_meta(self, file):
        file['artist'] = self.speaker_wav



from requests import post
from uuid import uuid4
from datetime import datetime
from io import BytesIO

from scipy.io.wavfile import read as read_wav

from .Raconteur import Raconteur


OAUTH_URL = 'https://ngw.devices.sberbank.ru:9443/api/v2/oauth'
TTS_URL = 'https://smartspeech.sber.ru/rest/v1/text:synthesize'

TIMEOUT = 100


class SaluteSpeech(Raconteur):
    name = 'salute'

    # def __init__(self, client_id: str, client_secret: str, auth: str, artist = 'Nec', *args, **kwargs):
    def __init__(self, auth: str, artist = 'Nec', *args, **kwargs):
        # self.client_id = client_id
        # self.client_secret = client_secret
        self.auth = auth
        self.artist = artist

        self.access_token = None
        self.access_token_expires = None

        super().__init__(*args, **kwargs)

    def predict(self, text: str):
        if self.access_token is None or datetime.now() > self.access_token_expires:
            self._refresh_access_token()

        response = post(
            TTS_URL,
            data = text.encode('utf-8'),
            params = {
                'format': 'wav16',
                'voice': f'{self.artist}_{self.sample_rate}'
            },
            headers = {
                'Authorization': f'Bearer {self.access_token}',
                'Content-Type': 'application/text'
            },
            verify = False,
            timeout = TIMEOUT
        )

        _, data = read_wav(BytesIO(response.content))

        # with open('audio.wav', 'wb') as file:
        #     file.write(response.content)

        # _, data = read_wav('audio.wav')
        # print(data.dtype)

        # write_wav('audioo.wav', self.sample_rate, data)

        return data

        # print(data)

    def _refresh_access_token(self):
        response = post(
            OAUTH_URL,
            data = {
                'scope': 'SALUTE_SPEECH_PERS'
            },
            headers = {
                'Authorization': f'Basic {self.auth}',
                'RqUID': str(uuid4())
            },
            verify = False,
            timeout = TIMEOUT
        ).json()

        self.access_token = response['access_token']
        self.access_token_expires = datetime.fromtimestamp(response['expires_at'] / 1000)

    @property
    def sample_rate(self):
        return 24_000

    @property
    def dtype(self):
        return 'int16'

    def set_file_meta(self, file):
        file['artist'] = self.artist



import os
from abc import ABC, abstractmethod
from uuid import uuid4

from scipy.io.wavfile import write as write_wav
import numpy as np
from IPython.display import Audio
from pydub import AudioSegment
from music_tag import load_file
from tqdm import tqdm

from .Splitter import Splitter
from .util import drop_empty_lines  # , drop_accent_marks


class Raconteur(ABC):
    name = None

    def __init__(self, splitter: Splitter, tmp_filename: str = None):
        if tmp_filename is None:
            self.tmp_filename = os.path.join(os.sep, 'tmp', f'{str(uuid4())}.wav')

        self.splitter = splitter

    def speak(
        self, text: str, filename: str = None, pbar: bool = False, save_text: str = True, accumulator = None,
        batch_size: int = None, first_batch_index: int = 0, title: str = None, update: bool = True
    ):
        audio = self._say(text, pbar = pbar, accumulator = accumulator, batch_size = batch_size, path_template = filename, first_batch_index = first_batch_index, title = title, update = update)

        if accumulator is None and audio is not None:
            self.save(audio, filename, text if save_text else None)

        return audio

    def save(self, audio: np.array, filename: str, text: str = None, title: str = None):
        write_wav(
            self.tmp_filename,
            self.sample_rate,
            audio
        )

        AudioSegment.from_wav(self.tmp_filename).export(filename, format = 'mp3')

        os.remove(self.tmp_filename)

        self.update(filename, text, title)

        return Audio(audio, rate = self.sample_rate)

    def update(self, filename: str, text: str = None, title: str = None):
        file = load_file(filename)

        if text is not None:
            file['title'] = text if title is None else title
            file['lyrics'] = text
            file['comment'] = text

            if title is not None:
                file['album'] = title[::-1].split('-', maxsplit = 1)[1][::-1]
            # file['year'] = 2023

        self.set_file_meta(file)

        file.save()

    @property
    @abstractmethod
    def sample_rate(self):
        pass

    @property
    @abstractmethod
    def dtype(self):
        pass

    @abstractmethod
    def predict(self, text):
        pass

    def set_file_meta(self, file):
        pass

    def _say(
        self, text: str, pbar: bool = False, accumulator = None, batch_size: int = None, path_template: str = None, first_batch_index: int = 0, title: str = None,
        update: bool = True
    ):
        if batch_size is None:
            combined = np.array([], dtype = self.dtype) if accumulator is None else accumulator

            items = self.splitter.split(text)
            chunks = tqdm(items) if pbar else items

            for chunk in chunks:
                # print(text)
                # print(chunk, len(chunk))

                combined = np.concatenate((combined, self.predict(chunk)))

            return combined

        assert accumulator is None, 'Accumulator must be none in a multibatch setting'
        assert path_template is not None, 'Path template is required in a multibatch setting'
        assert title is not None, 'Title template is required in a multibatch setting'

        combined = np.array([], dtype = self.dtype)

        items = self.splitter.split(text)
        chunks = tqdm(items) if pbar else items

        batch_length = 0
        batch_index = first_batch_index
        text = ''

        def save():
            nonlocal batch_length, combined, batch_index, text

            if update:
                self.update(path_template.format(batch = batch_index), drop_empty_lines(text), title = title.format(batch = batch_index))
                # self.update(path_template.format(batch = batch_index), drop_empty_lines(drop_accent_marks(text)), title.format(batch = batch_index))
                # self.update(path_template.format(batch = batch_index), drop_empty_lines(drop_accent_marks(text)), f'{batch_index:02d}')
            else:
                # self.save(combined, path_template.format(batch = batch_index), drop_empty_lines(drop_accent_marks(text)), title = title.format(batch = batch_index))
                self.save(combined, path_template.format(batch = batch_index), drop_empty_lines(text), title = title.format(batch = batch_index))

            batch_length = 0
            combined = np.array([], dtype = self.dtype)
            batch_index += 1
            text = ''

        for chunk in chunks:
            # print(text)
            # print(chunk, len(chunk))

            text = f'{text} {chunk}'
            if not update:
                combined = np.concatenate((combined, self.predict(chunk)))

            batch_length += len(chunk)

            if batch_length >= batch_size:
                save()

        if combined.shape[0] > 0 or update and os.path.isfile(path_template.format(batch = batch_index)):
            save()

        return None



import re

from num2words import num2words


NUMBER_PATTERN = re.compile('[-+]?[0-9]+')

NEW_LINE = '\n'
DOUBLE_NEW_LINE = re.compile('\n\n')

# ACCENT_MARK_OCCURRENCE = re.compile(r'([^\s])\+([^\s])')
PLUS = '+'


def translate_numbers(text: str, lang: str = 'ru'):
    while (match := NUMBER_PATTERN.search(text)) is not None:
        number = match.group(0)
        text = text.replace(number, num2words(number, lang = lang), 1)

    return text


def drop_empty_lines(text: str):
    return DOUBLE_NEW_LINE.sub(NEW_LINE, text)


def drop_accent_marks(text: str):
    return text.replace(PLUS, '')



# Copied from https://github.com/zeionara/marude/blob/master/marude/cli/SpeechIndex.py

from os import walk, path
from pathlib import Path
from collections import namedtuple

Key = namedtuple('AnekId', field_names = ('source', 'id'))
Location = namedtuple('Location', field_names = ('path', 'file'))


class SpeechIndex:
    def __init__(self, root: str):
        self.root = root

        index = {}

        n_files = 0

        for directory, _, files in walk(root):
            for file in files:
                components = Path(file).stem.split('.')

                assert len(components) == 2, f'Incorrect file name {file}'

                index[Key(source = components[1], id = int(components[0]))] = Location(path = path.join(directory, file), file = file)

                n_files += 1

        print(f'Indexed {n_files} files')

        self._index = index

    def get(self, source: str, id_: int):
        return self._index[Key(source = source, id = id_)]



import time
from os import environ as env

env['PYGAME_HIDE_SUPPORT_PROMPT'] = 'hide'

import pygame
import numpy as np


N_MILLISECONDS_IN_SECOND = 1000


class Beep():
    def __init__(
        self, n_repetitions: int = 3, sample_rate = 40100, bit_depth = -16, n_channels = 1, duration = 1000,
        error_signal = 10, error_frequency = 5000, error_fraction = 0.9, error_n_duplicates = 5, error_pause_duration = 0.5,
        ok_signal = 1000, ok_fraction = 0.9
    ):
        self.n_repetitions = n_repetitions
        self.sample_rate = sample_rate
        self.bit_depth = bit_depth
        self.n_channels = n_channels
        self.duration = duration

        self.error_signal = error_signal
        self.error_frequency = error_frequency
        self.error_fraction = error_fraction
        self.error_n_duplicates = error_n_duplicates
        self.error_pause_duration = error_pause_duration

        self.ok_signal = ok_signal
        self.ok_fraction = ok_fraction

    def __enter__(self):
        pygame.mixer.init()

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if exception_type:
            for _ in range(self.n_repetitions):
                pygame.mixer.init(frequency = self.sample_rate, size = self.bit_depth, channels = self.n_channels)

                values = np.array([
                    item for item in np.linspace(self.error_signal * self.error_fraction, self.error_signal, self.error_frequency)
                    for _ in range(self.error_n_duplicates)
                ])

                sound = pygame.mixer.Sound(buffer = values.tobytes())

                sound.play()

                time.sleep(self.duration * (1 + self.error_pause_duration) / N_MILLISECONDS_IN_SECOND)
        else:
            pygame.mixer.init(frequency = self.sample_rate, size = self.bit_depth, channels = self.n_channels)

            values = np.linspace(self.ok_signal * self.ok_fraction, self.ok_signal, int(self.sample_rate * self.duration / N_MILLISECONDS_IN_SECOND))

            sound = pygame.mixer.Sound(buffer = values.tobytes())

            sound.play()

            time.sleep(self.duration / N_MILLISECONDS_IN_SECOND)


beep = Beep



from bark import SAMPLE_RATE, generate_audio

from .Raconteur import Raconteur


class Bark(Raconteur):
    name = 'bark'

    def __init__(self, artist: str, *args, **kwargs):
        self.artist = artist
        super().__init__(*args, **kwargs)

    @property
    def sample_rate(self):
        return SAMPLE_RATE

    @property
    def dtype(self):
        return 'float32'

    def predict(self, text):
        return generate_audio(text, history_prompt = self.artist)

    def set_file_meta(self, file):
        file['artist'] = self.artist



import base64
from uuid import uuid4
from io import BytesIO

from requests import post
from scipy.io.wavfile import read as read_wav

from .Raconteur import Raconteur


SESSION_MANAGEMENT_URL = 'https://cloud.speechpro.com/vksession/rest/session'
SYNTHESIS_URL = 'https://cloud.speechpro.com/vktts/rest/v1/synthesize'

TIMEOUT = 100


class Crt(Raconteur):
    name = 'crt'

    def __init__(self, username: str, password: str, domain: int, artist: str, *args, **kwargs):
        self.username = username
        self.password = password
        self.domain = domain

        self.artist = artist

        self.session_id = None

        super().__init__(*args, **kwargs)

    def predict(self, text: str):
        if self.session_id is None:
            self._create_session()

        response = post(
            SYNTHESIS_URL,
            headers = {
                'X-Session-ID': self.session_id,
                'X-Request-Id': str(uuid4())
            },
            json = {
                'voice_name': self.artist,
                'text': {
                    'mime': 'text/plain',
                    'value': text
                },
                'audio': 'audio/wav'
            },
            timeout = TIMEOUT
        )

        data = response.json().get('data')

        if data is None:
            raise KeyError(f'No "data" field in response from crt service: {response.text}. Probably, the quota has been exhausted')

        _, data = read_wav(
            BytesIO(
                base64.decodebytes(
                    bytes(
                        data,
                        'utf-8'
                    )
                )
            )
        )

        return data

    @property
    def sample_rate(self):
        return 22_050

    @property
    def dtype(self):
        return 'int16'

    def _create_session(self):

        response = post(
            SESSION_MANAGEMENT_URL,
            json = {
                'username': self.username,
                'password': self.password,
                'domain_id': self.domain
            },
            timeout = TIMEOUT
        )

        self.session_id = response.json()['session_id']

    def set_file_meta(self, file):
        file['artist'] = self.artist



from setuptools import setup, find_packages


def get_readme():
    with open('README.md', 'r', encoding = 'utf-8') as file:
        return file.read()


setup(
    name = 'raconteur',
    packages = find_packages(),
    version = '0.5',
    license = 'Apache 2.0',
    description = 'An auxiliary app for simplifying speech synthesis on arbitrary texts',
    long_description = get_readme(),
    author = 'Zeio Nara',
    author_email = 'zeionara@gmail.com',
    url = 'https://github.com/zeionara/raconteur',
    keywords = ['speech', 'synthesis'],
    install_requires = [
        'ipython', 'scipy', 'pydub', 'music-tag', 'numpy', 'transliterate',
        'suno-bark @ git+https://github.com/suno-ai/bark.git',
        'RUTTS @ git+https://github.com/zeionara/RUTTS'
    ],
    classifiers = [
        'Programming Language :: Python :: 3.11'
    ]
)
